/****************************************************************************
Filename: keypad.c 
Version: 1.00
Description: scan keypad input using RTC periodic IRQ output as pacing clock
By: M. Johnson
****************************************************************************
Date: 		Revision:	By:		Description:

20010612	1.00		mtj		Original
20020603	1.10		mtj		Added support for TS-3200 hardware

****************************************************************************/
/*

the IRQ output on the RTC is connected to IRQ8 on the microprocessor through 
an inverter. This example uses the periodic interrupt feature of the RTC to 
call our interrupt service routine (ISR) at a 256 Hz rate.The ISR takes the 
row outputs low and scans the column inputs for a keypress. The kepress is
decoded and the result is displayed on the standard output device, which could
be COM2, video, LCD, etc (default is COM2). 

For our test we used a TS-2200 embedded PC connected to a Grayhill Series 96
4x4 keypad using digital I/O port #2. The keypad looks just like an american
telephone keypad except it has an additional row of A-D on the right side.

columns are connected to digital inputs and pulled up to +5v with 10k ohms.
		 

| | | |

1 2 3 A -
4 5 6 B -   rows are connected to digital outputs.
7 8 9 C -
* 0 # D -


Connection Table for TS-2200

DIO #2 pin	DIO name	keypad		keypad pin
==========	========	======		==========

pin 8 		 P1.1		row 1-A		pin 1
pin 6 		 P1.2		row 4-B		pin 2
pin 14		 P1.3		row 7-C		pin 3
pin 13		 P3.2		row *-D		pin 4
  
pin 7 		inp 1		col 1-*		pin 5 
pin 5 		inp 2		col 2-0		pin 6
pin 3 		inp 3		col 3-#		pin 7 
pin 1 		inp 4		col A-D		pin 8



Connection Table for TS-3200

DIO #2 pin	DIO name	keypad		keypad pin
==========	========	======		==========

pin 1		DIO2.0		row 1-A		pin 1
pin 3		DIO2.1		row 4-B		pin 2
pin 5		DIO2.2		row 7-C		pin 3
pin 7		DIO2.3		row *-D		pin 4
  
pin 9		DIO2.4		col 1-*		pin 5 
pin 11		DIO2.5		col 2-0		pin 6
pin 13		DIO2.6		col 3-#		pin 7 
pin 15		DIO2.7		col A-D		pin 8



(link to RTC datasheet) http://www.embeddedx86.com/epc/support/12887.pdf

DS12887,
0x70 selects which register within the DS12887 to access, write only.
0x71 provides access to reg selected in 0x70, read/write.

info@embeddedx86.com

Technlogic Systems, Inc.  www.embeddedx86.com

*/               


/****************************************************************************
	INCLUDE FILES
****************************************************************************/

#include <dos.h>
#include <stdio.h>

/****************************************************************************
	TYPEDEFS AND DEFINES
****************************************************************************/

#define TS2200	0
#define TS3200	1

#define FALSE		0
#define TRUE		!FALSE

#define DEBUG		0			// set to zero to disable debug printf calls
#define LED			1			// set to zero to disable LED (D2) toggle

// register definitions for 12887 Real Time Clock 

#define RTC_CMD_ADDR	0x70	// RTC internal register offset goes here 
#define RTC_DAT_ADDR	0x71	// RTC internal register R/W access here

#define RTC_REG_A		0x0A	// RTC register offset for accessing Reg. A
#define RTC_REG_B		0x0B	// RTC register offset for accessing Reg. B
#define RTC_REG_C		0x0C	// RTC register offset for accessing Reg. C
#define RTC_REG_D		0x0D	// RTC register offset for accessing Reg. D

// register definitions for 8259 programmable interrupt controller

#define INT_0_CTRL_ADDR	0x20	// 8259 master (0) interrupt control register
#define INT_0_MASK_ADDR	0x21	// 8259 master (0) interrupt mask register
#define INT_1_CTRL_ADDR	0xA0	// 8259 slave  (1) interrupt control register
#define INT_1_MASK_ADDR	0xA1	// 8259 slave  (1) interrupt mask register   

#define END_OF_INT		0x20	// generic end of interrupt mask

// defines for TS-3200

#define DIO2_CTRL	0x7D		// control byte for DIO2 and LCD ports
#define DIO2_DATA	0x7E		// data latch for DIO2

// enum for keypad state machine

enum _eKeyState {
	WAITING,
	KEY_PRESSED,
	KEY_DECODED
};

/****************************************************************************
	GLOBAL VARIABLES
****************************************************************************/

#if DEBUG
void interrupt(*NewVector)();		// holds our vector address, for debug
#endif

void interrupt(*OldVector)();		// holds the previous vector for INT8

volatile unsigned char uc_RTC_A;	// storage for RTC reg. A
volatile unsigned char uc_RTC_B;	// storage for RTC reg. B
volatile unsigned char uc_RTC_C;	// storage for RTC reg. C
volatile unsigned char uc_RTC_D;	// storage for RTC reg. D

volatile unsigned char ucPort;		// temp storage for port I/O
volatile unsigned char ucCount;		// generic counter storage

volatile unsigned char ucColumn[4];	// stores result of column reads

char cKeyMatrix[4][4] = {   '1', '2', '3', 'A',		// keypad decode map
	'4', '5', '6', 'B',
	'7', '8', '9', 'C',
	'*', '0', '#', 'D'};

unsigned char i,j;					// generic counter variables

enum _eKeyState eKeyState;			// enum for tracking keypad state

unsigned char ucKeyPressed;			// flag to track when key is being pressed

char cKey;							// stores the pressed key

union REGS regs;	// allows register access, used to call BIOS interrupts

/****************************************************************************
	FUNCTION PROTOTYPES
****************************************************************************/

void StartKeypadScan( void );
void StopKeypadScan( void );
void interrupt RealTimeClockISR( void );
void ScanKeypad( void );

/****************************************************************************
	EXTERNAL DECLARATIONS
****************************************************************************/

/****************************************************************************
Function: main
Description: initialization, main loop, de-initialization
Passed: none
Returns: none
*****************************************************************************/

main()
{

	eKeyState = WAITING;				// start out waiting for keypress

#if TS2200

	//-------------------------------------------------------------------------
	// Port 1, bits 1-3 drive the scan rows, make them output, CMOS, high
	//-------------------------------------------------------------------------

	ucPort = inport( 0xF820 );			// read port 1 mode config. register

	outport( 0xF820, (ucPort & 0xF1) );	// set  port 1 bits 1-3 as I/O

	ucPort = inport( 0xF864 );			// read port 1 direction register

	outport( 0xF864, (ucPort & 0xF1) );	// set  port 1 bits 1-3 as CMOS outputs

	ucPort = inport( 0xF862 );			// read port 1 latch register

	outport( 0xF862, (ucPort | 0x0E) );	// set  port 1 bits 1-3 high

	//-------------------------------------------------------------------------
	// Port 3, bit 4 drives the last scan row, make it output, CMOS, high
	//-------------------------------------------------------------------------

	ucPort = inport( 0xF824 );			// read port 3 mode config. register

	outport( 0xF824, (ucPort & 0xEF) );	// set  port 3 bit 4 as I/O

	ucPort = inport( 0xF874 );			// read port 3 direction register

	outport( 0xF874, (ucPort & 0xEF) );	// set  port 3 bit 4 as CMOS outputs

	ucPort = inport( 0xF872 );			// read port 3 latch register

	outport( 0xF872, (ucPort | 0x10) );	// set  port 3 bit 4 high

#endif

#if TS3200

	// set LCD port to LCD mode, DIO2 0-3 inputs, DIO2 4-7 outputs

	outportb( DIO2_CTRL, 0x02 );

#endif


	//-------------------------------------------------------------------------
	// Initialize RTC IRQ output, setup ISR vector, enable IRQ8
	//-------------------------------------------------------------------------

	StartKeypadScan();

	//-------------------------------------------------------------------------
	// Dump vectors and RTC registers if debug is enabled
	//-------------------------------------------------------------------------

#if DEBUG
	NewVector = getvect(0x70);					// get int. vector for RTC (IRQ8)

	printf( "OldVector=(%p)\n", OldVector );
	printf( "NewVector=(%p)\n", NewVector );
#endif

#if DEBUG
	outportb( RTC_CMD_ADDR, RTC_REG_A );	// select RTC register A 
	uc_RTC_A = inportb( RTC_DAT_ADDR );		// read   RTC register A 

	outportb( RTC_CMD_ADDR, RTC_REG_B );	// select RTC register B 
	uc_RTC_B = inportb( RTC_DAT_ADDR );		// read   RTC register B 

	outportb( RTC_CMD_ADDR, RTC_REG_C );	// select RTC register C 
	uc_RTC_C = inportb( RTC_DAT_ADDR );		// read   RTC register C 

	outportb( RTC_CMD_ADDR, RTC_REG_D );	// select RTC register D 
	uc_RTC_D = inportb( RTC_DAT_ADDR );		// read   RTC register D 

	printf( "A=(%02X)\n", uc_RTC_A );
	printf( "B=(%02X)\n", uc_RTC_B );
	printf( "C=(%02X)\n", uc_RTC_C );
	printf( "D=(%02X)\n", uc_RTC_D );
#endif

	//-------------------------------------------------------------------------
	//  main loop 
	//-------------------------------------------------------------------------

	while( !kbhit() ) {						// exit on any PC keyboard keypress

		if( eKeyState == KEY_DECODED ) {	// if keypad input decoded
			printf( "%c", cKey );			// display keypress to standard output
			eKeyState = WAITING;			// reset state to waiting for keypress
		}

// This is debug stuff you can uncomment if things don't work.

//		printf( "0=%02X "  , ucColumn[0] );
//		printf( "1=%02X "  , ucColumn[1] );
//		printf( "2=%02X "  , ucColumn[2] );
//		printf( "3=%02X "  , ucColumn[3] );

/*		switch( eKeyState )
			{
			case WAITING:
				printf("WAITING\n");
				break;
			case KEY_PRESSED:
				printf("KEY PRESSED\n");
				break;
			case KEY_DECODED:
				printf("KEY DECODED\n");
				break;
			default :
				printf("no match\n");
				break;
			}
*/
	};

	//-------------------------------------------------------------------------
	//  end main loop, stop RTC interrupts, disable IRQ8, restore old vector
	//-------------------------------------------------------------------------

	StopKeypadScan();                           

}

/****************************************************************************
Function: StartKeypadScan
Description: Setup our ISR vector, enable interrupts
Passed: none
Returns: none
*****************************************************************************/

void StartKeypadScan( void )
{

	// IRQ8 calls the interrupt service routine installed at vector 0x70

	//-------------------------------------------------------------------------
	// Setup interrupt vectors so that IRQ8 calls our interrupt service routine
	//-------------------------------------------------------------------------

	OldVector = getvect(0x70);			// get old int. vector to re-install on exit

	disable();							// disable interrupts for next call

	setvect(0x70,RealTimeClockISR);		// set RTC int. vector for our routine

	enable();							// enable system interrupts

	//-------------------------------------------------------------------------
	// setup RTC to generate periodic interrupts on IRQ8 @ 256hz rate (3.906mS)
	//-------------------------------------------------------------------------

	disable();	// don't let interrupts happen during write to RTC registers

	outportb( RTC_CMD_ADDR, RTC_REG_B );	// select RTC register B
	outportb( RTC_DAT_ADDR, 0x42 );			// set Periodic Interrupt Enable bit		

	outportb( RTC_CMD_ADDR, RTC_REG_A );	// select RTC register A
	outportb( RTC_DAT_ADDR, 0x28 );			// set rate to 256hz, oscillator enabled

	enable();	// re-enable system interrupts

	//-------------------------------------------------------------------------
	// setup interrupt controller to enable IRQ8 (interrupt 0 on second 8259)
	//-------------------------------------------------------------------------

	disable();	// don't let interrupts happen during read/modify/write

	ucPort = inportb( INT_1_MASK_ADDR );	// read interrupt mask, 8259(1)

	ucPort &= 0xFE;							// clear mask bit for IRQ8

	outportb( INT_1_MASK_ADDR, ucPort );	// write modified mask 8259(1)

	enable();	// re-enable system interrupts
}

/****************************************************************************
Function: StopKeypadScan
Description: disable our interrupts, reset old vector
Passed: none
Returns: none
*****************************************************************************/

void StopKeypadScan( void )
{

	//-------------------------------------------------------------------------
	// setup interrupt controller to disable IRQ8 (interrupt 0 on second 8259)
	//-------------------------------------------------------------------------

	disable(); // don't let interrupts happen during read/modify/write

	ucPort = inportb( INT_1_MASK_ADDR );	// read interrupt mask, 8259(1)

	ucPort |= 0x01;							// clear mask bit for IRQ8

	outportb( INT_1_MASK_ADDR, ucPort );	// write modified mask 8259(1)

	enable();  // re-enable system interrupts

	//-------------------------------------------------------------------------
	// setup RTC to disable periodic interrupts on IRQ8
	//-------------------------------------------------------------------------

	disable(); // don't let interrupts happen during write to RTC registers

	outportb( RTC_CMD_ADDR, RTC_REG_B );	// select RTC register B
	outportb( RTC_DAT_ADDR, 0x02 );			// clear Periodic Interrupt Enable bit

	outportb( RTC_CMD_ADDR, RTC_REG_A );	// select RTC register A
	outportb( RTC_DAT_ADDR, 0x28 );			// set rate to 256hz, oscillator disabled

	enable();  // re-enable system interrupts

	//-------------------------------------------------------------------------
	// Setup interrupt vectors so that IRQ8 calls our interrupt service routine
	//-------------------------------------------------------------------------

	disable();								// disable interrupts for next call

	setvect( 0x70, OldVector );				// set RTC int. vector for our routine

	enable();								// enable system interrupts

	//-------------------------------------------------------------------------
	// turn off LED
	//-------------------------------------------------------------------------

#if LED

	// use the BIOS calls so this code will work on all products.

	regs.x.ax = 0xB010;
	regs.h.bh = 0x00;
	int86 (0x15, &regs, &regs);
#endif

}

/****************************************************************************
Function: RealTimeClockISR
Description: interrupt service routine, resets flags, calls scan routine.
Passed: none
Returns: none
*****************************************************************************/

void interrupt RealTimeClockISR( void )
{

	//-------------------------------------------------------------------------
	// reading from register C on the DS12887 RTC will clear the interrupt
	//-------------------------------------------------------------------------

	outportb( RTC_CMD_ADDR, RTC_REG_C );	// select RTC register C
	uc_RTC_C = inportb( RTC_DAT_ADDR );		// read   RTC register C

	//-------------------------------------------------------------------------
	// send end of interrupt to second, then first 8259 interrupt controller
	//-------------------------------------------------------------------------

	outportb( INT_1_CTRL_ADDR, END_OF_INT );
	outportb( INT_0_CTRL_ADDR, END_OF_INT );

	//-------------------------------------------------------------------------
	// toggle LED D2 once every 255 interrupts (255*3.906mSec) = 1 second
	//-------------------------------------------------------------------------

// use the BIOS calls so this code will work on all products.

#if LED
	if( ucCount++ > 254 ) {
		ucCount = 0;

		// toggle LED
		regs.x.ax = 0xB010;
		regs.h.bh = 0x81;
		int86 (0x15, &regs, &regs);
	}

#endif

	ScanKeypad();

}

/****************************************************************************
Function: ScanKeypad
Description: Takes each row output low and stores column input, tests for keypress
Passed: none
Returns: none
*****************************************************************************/

void ScanKeypad( void )
{

// if a key press is not stored and ready for decode, scan keypad

	if( eKeyState != KEY_DECODED ) {

#if TS2200
		ucPort = inport( 0xF862 );			// read port 1 latch register

		outport( 0xF862, (ucPort & 0xFD) );	// set  port 1 bit 1 low

		ucColumn[0] = inport( 0x77 );		// read column inputs

		outport( 0xF862, (ucPort | 0x02) );	// set  port 1 bit 1 high


		ucPort = inport( 0xF862 );			// read port 1 latch register

		outport( 0xF862, (ucPort & 0xFB) );	// set  port 1 bit 2 low

		ucColumn[1] = inport( 0x77 );		// read column inputs

		outport( 0xF862, (ucPort | 0x04) );	// set  port 1 bit 2 high


		ucPort = inport( 0xF862 );			// read port 1 latch register

		outport( 0xF862, (ucPort & 0xF7) );	// set  port 1 bit 3 low

		ucColumn[2] = inport( 0x77 );		// read column inputs

		outport( 0xF862, (ucPort | 0x08) );	// set  port 1 bit 3 high


		ucPort = inport( 0xF872 );			// read port 3 latch register

		outport( 0xF872, (ucPort & 0xEF) );	// set  port 3 bit 4 low

		ucColumn[3] = inport( 0x77 );		// read column inputs

		outport( 0xF872, (ucPort | 0x10) );	// set  port 3 bit 4 high
#endif

#if TS3200
		ucPort = inportb( DIO2_DATA );			// read port 1 latch register

		outportb( DIO2_DATA,(ucPort & 0xEF));	// set DIO2 bit 4 low

		ucColumn[0] = inportb( DIO2_DATA );		// read column inputs

		outport( DIO2_DATA, (ucPort | 0x10) );	// set DIO2 bit 4 high


		ucPort = inportb( DIO2_DATA );			// read port 1 latch register

		outportb( DIO2_DATA,(ucPort & 0xDF));	// set DIO2 bit 5 low

		ucColumn[1] = inportb( DIO2_DATA );		// read column inputs

		outport( DIO2_DATA, (ucPort | 0x20) );	// set DIO2 bit 5 high


		ucPort = inportb( DIO2_DATA );			// read port 1 latch register

		outportb( DIO2_DATA,(ucPort & 0xBF));	// set DIO2 bit 6 low

		ucColumn[2] = inportb( DIO2_DATA );		// read column inputs

		outport( DIO2_DATA, (ucPort | 0x40) );	// set DIO2 bit 6 high


		ucPort = inportb( DIO2_DATA );			// read port 1 latch register

		outportb( DIO2_DATA,(ucPort & 0x7F));	// set DIO2 bit 7 low

		ucColumn[3] = inportb( DIO2_DATA );		// read column inputs

		outport( DIO2_DATA, (ucPort | 0x80) );	// set DIO2 bit 7 high
#endif

	}

	ucKeyPressed = FALSE;					// clear KeyPressed flag


	for( i=0; i<4; i++ )
#if TS2200
		if( (ucColumn[i] & 0x1E) ^ 0x1E )	// check for key being pressed
#endif
#if TS3200
		if( (ucColumn[i] & 0x0F) ^ 0x0F )
#endif
			ucKeyPressed = TRUE;			// set flag if key is pressed

		// if the current state is WAITING and a key is pressed, store the key and
		// wait for the user to release the key.

	if( ucKeyPressed && (eKeyState == WAITING) ) {
		eKeyState = KEY_PRESSED;

		// store keypress here

		for( i=0; i<4; i++ ) {
			for( j=0; j<4; j++ ) {
#if TS2200
				if( (~ucColumn[i]) & (0x02 << j) ) {
					cKey = cKeyMatrix[i][j];
#endif
#if TS3200
					if( (~ucColumn[i]) & (0x01 << j) ) {
						cKey = cKeyMatrix[j][i];
#endif
					}
				}
			}
		}

		// if the current state is KEY_PRESSED and a key is not being pressed,
		// the user has released the key. Time to decode the key press stored above.

		if( !ucKeyPressed && (eKeyState == KEY_PRESSED) )
			eKeyState = KEY_DECODED;

		//-------------------------------------------------------------------------
		//-------------------------------------------------------------------------

	}

